@@ -14,6 +14,12 @@
 #include <errno.h>
 #include <syslog.h>
 
+#ifdef PERL_SET_CONTEXT
+#define PSC(x) PERL_SET_CONTEXT(x)
+#else
+#define PSC(x) (void) 0
+#endif
+
 #define PERLPARSE_NUM_ARGS 6
 
 static PerlInterpreter *my_perl = NULL;
@@ -31,6 +37,8 @@ void
 term_embedded_interpreter(void)
 {
     if (my_perl != NULL) {
+	PSC(my_perl);
+	PERL_SET_INTERP(my_perl);
 	PL_perl_destruct_level = 1;
 	perl_destruct(my_perl);
 	perl_free(my_perl);
@@ -67,6 +75,8 @@ make_embedded_interpreter(char const *progPath,
 
     if (my_perl != NULL) {
 #ifdef SAFE_EMBED_PERL
+	PSC(my_perl);
+	PERL_SET_INTERP(my_perl);
 	PL_perl_destruct_level = 1;
 	perl_destruct(my_perl);
 	perl_free(my_perl);
@@ -107,9 +117,8 @@ make_embedded_interpreter(char const *progPath,
 	errno = ENOMEM;
 	return -1;
     }
-#ifdef PERL_SET_CONTEXT
-    PERL_SET_CONTEXT(my_perl);
-#endif
+    PSC(my_perl);
+    PERL_SET_INTERP(my_perl);
     PL_perl_destruct_level = 1;
     perl_construct(my_perl);
     perl_parse(my_perl, xs_init, argc, argv, NULL);
@@ -135,6 +144,8 @@ run_embedded_filter(void)
 {
     char *args[] = { NULL };
 
+    PSC(my_perl);
+    PERL_SET_INTERP(my_perl);
     embperl_fix_pid();
 
     perl_call_argv("do_main_loop", G_DISCARD | G_NOARGS, args);
