@@ -14,9 +14,6 @@
 *
 ***********************************************************************/
 
-static char const RCSID[] =
-"$Id$";
-
 #include "event.h"
 #include <stdlib.h>
 #include <errno.h>
@@ -76,6 +73,28 @@ set_cloexec(int fd)
     }
 }
 
+/**********************************************************************
+* %FUNCTION: set_nonblocking
+* %ARGUMENTS:
+*  fd -- file descriptor
+* %RETURNS:
+*  0 on success, -1 on failure
+* %DESCRIPTION:
+*  Sets the O_NONBLOCK flag on fd
+***********************************************************************/
+int
+set_nonblocking(int fd)
+{
+    int flags = fcntl(fd, F_GETFL, 0);
+    if (flags < 0) return flags;
+    if (!(flags & O_NONBLOCK)) {
+	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
+	    return -1;
+	}
+    }
+    return 0;
+}
+
 /**********************************************************************
 * %FUNCTION: Event_CreateSelector
 * %ARGUMENTS:
@@ -403,7 +422,7 @@ Event_HandleEventUsingPoll(EventSelector *es)
 	/* Go through poll set and set pollflags */
 	for (i=0; i<nfds; i++) {
 	    eh = Fd_to_eh[i];
-	    if (Pollfds[i].revents & POLLIN) {
+	    if (Pollfds[i].revents & (POLLIN|POLLHUP)) {
 		if (eh->flags & EVENT_FLAG_READABLE) {
 		    eh->pollflags |= EVENT_FLAG_READABLE;
 		}
