@@ -7,8 +7,6 @@
 *
 * Copyright (C) 2001-2003 Roaring Penguin Software Inc.
 *
-* $Id$
-*
 * This program may be distributed according to the terms of the GNU
 * General Public License, version 2 or (at your option) any later version.
 *
@@ -39,6 +37,7 @@ typedef void (*EventTcpIOFinishedFunc)(EventSelector *es,
 				       int flag,
 				       void *data);
 
+#define EVENT_TCP_FLAG_INIT     -1
 #define EVENT_TCP_FLAG_COMPLETE 0
 #define EVENT_TCP_FLAG_IOERROR  1
 #define EVENT_TCP_FLAG_EOF      2
@@ -51,6 +50,7 @@ typedef struct EventTcpState_t {
     int len;
     int chunked;
     int delim;
+    int got_delim;
     EventTcpIOFinishedFunc f;
     EventSelector *es;
     EventHandler *eh;
