@@ -151,9 +151,9 @@ static char *EquivToLoopback = NULL;
 static char ValidateHeader[256];
 
 /* Additional Sendmail macros to pass along */
-#define MAX_ADDITIONAL_SENDMAIL_MACROS 256
-static char *AdditionalMacros[MAX_ADDITIONAL_SENDMAIL_MACROS];
-static int NumAdditionalMacros = 0;
+/* Note that libmilter has a hard limit of 50
+   and we include 20 in StandardSendmailMacros */
+#define MAX_ADDITIONAL_SENDMAIL_MACROS 30
 
 /* Standard Sendmail macros */
 /* We can't make it char const * because libmilter
@@ -161,17 +161,23 @@ static int NumAdditionalMacros = 0;
 static char *StandardSendmailMacros[] = {
     "_", "auth_authen", "auth_author", "auth_ssf", "auth_type",
     "cert_issuer", "cert_subject", "cipher", "cipher_bits", "daemon_name",
+    "daemon_port",
     "i", "if_addr", "if_name", "j", "mail_addr", "mail_host", "mail_mailer",
     "tls_version", "verify",
     /* End of macros MUST be marked with NULL! */
     NULL
 };
 
+static char *AdditionalMacros[MAX_ADDITIONAL_SENDMAIL_MACROS];
+static int NumAdditionalMacros = 0;
+
 /* Keep track of private data -- file name and fp for writing e-mail body */
 struct privdata {
     char *hostname;		/* Name of connecting host */
     char *hostip;		/* IP address of connecting host */
+    unsigned int hostport;      /* Port of connecting host */
     char *myip;                 /* My IP address, from Sendmail macro */
+    unsigned int daemon_port;   /* Daemon port from Sendmail macro */
     char *sender;		/* Envelope sender */
     char *firstRecip;		/* Address of first recipient */
     char *dir;			/* Work directory */
@@ -401,7 +407,14 @@ set_reply(SMFICTX *ctx,
 	  char const *reply)
 {
     char *safe_reply;
-    sfsistat r;
+
+    char stack_safe_reply[SMALLBUF];
+
+    sfsistat retcode;
+    char const *s;
+    char *t;
+    int len;
+
     if (!reply || !*reply) {
 	if (*first == '4') {
 	    reply = "Please try again later";
@@ -419,50 +432,47 @@ set_reply(SMFICTX *ctx,
     }
 
     /* We need to double any "%" chars in reply */
-    if (strchr(reply, '%')) {
-	sfsistat retcode;
-	char const *s;
-	char *t;
-	/* Worst-case, we'll double our length */
-	safe_reply = malloc(strlen(reply) * 2 + 1);
+    /* Worst-case, we'll double our length */
+    len = strlen(reply) * 2 + 1;
+    if (len <= SMALLBUF) {
+	/* avoid malloc if the reply is short enough */
+	safe_reply = stack_safe_reply;
+    } else {
+	safe_reply = malloc(len);
 	if (!safe_reply) {
 	    syslog(LOG_ERR, "Out of memory to escape reply %s", reply);
 	    return smfi_setreply(ctx, (char *) code, (char *) dsn, "Out of memory");
 	}
-	s = reply;
-	t = safe_reply;
-	while (*s) {
-	    if (*s == '%') *t++ = '%';
+    }
+
+    s = reply;
+    t = safe_reply;
+    while (*s) {
+	if (*s == '%') *t++ = '%';
+	/* Replace control chars except for \n with a space */
+	if ((*s < ' ' && *s != '\n') || *s == 0x7f) {
+	    *t++ = ' ';
+	    s++;
+	} else {
 	    *t++ = *s++;
 	}
-	*t = 0;
-	if (!strchr(reply, '\n')) {
-	    retcode = smfi_setreply(ctx, (char *) code, (char *) dsn, safe_reply);
-	} else {
-	    retcode = do_reply(ctx, code, dsn, safe_reply);
-	}
-	free(safe_reply);
-	return retcode;
     }
-
-    /* If there are no carriage-returns, avoid allocating memory */
+    *t = 0;
     if (!strchr(reply, '\n')) {
-	return smfi_setreply(ctx, (char *) code, (char *) dsn, (char *) reply);
+	retcode = smfi_setreply(ctx, (char *) code, (char *) dsn, safe_reply);
+    } else {
+	retcode = do_reply(ctx, code, dsn, safe_reply);
     }
-
-    safe_reply = strdup_with_log(reply);
-    if (!safe_reply) {
-	return smfi_setreply(ctx, (char *) code, (char *) dsn, "Out of memory");
+    if (safe_reply != stack_safe_reply) {
+	free(safe_reply);
     }
-    r = do_reply(ctx, code, dsn, safe_reply);
-    free(safe_reply);
-    return r;
+    return retcode;
 }
 
 /**********************************************************************
 *%FUNCTION: closefiles
 *%ARGUMENTS:
-* None
+* notthis - file descriptor that should not be closed
 *%RETURNS:
 * Nothing
 *%DESCRIPTION:
@@ -470,11 +480,11 @@ set_reply(SMFICTX *ctx,
 *         to conserve descriptors.
 ***********************************************************************/
 static void
-closefiles(void)
+closefiles(int notthis)
 {
     int i;
     for (i=0; i<CLOSEFDS; i++) {
-	(void) close(i);
+	if (i != notthis) (void) close(i);
     }
 }
 
@@ -495,6 +505,7 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     struct privdata *data;
 
     char const *tmp;
+    char *me;
     struct sockaddr_in *insa = (struct sockaddr_in *) sa;
 #if defined(AF_INET6) && defined(HAVE_INET_NTOP)
     struct sockaddr_in6 *in6sa = (struct sockaddr_in6 *) sa;
@@ -550,7 +561,9 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     }
     data->hostname = NULL;
     data->hostip   = NULL;
+    data->hostport = 0;
     data->myip     = NULL;
+    data->daemon_port = 0;
     data->sender   = NULL;
     data->firstRecip = NULL;
     data->dir      = NULL;
@@ -587,6 +600,17 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
 	data->hostname = NULL;
     }
 
+    /* Port */
+    if (sa) {
+#ifdef AF_INET6
+      if (sa->sa_family == AF_INET6) {
+	data->hostport = ntohs(in6sa->sin6_port);
+      }
+#endif
+      if (sa->sa_family == AF_INET) {
+	data->hostport = ntohs(insa->sin_port);
+      }
+    }
     /* Padding -- should be big enough for IPv6 addresses */
     if (!sa) {
 	data->hostip = NULL;
@@ -661,10 +685,30 @@ mfconnect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *sa)
     data->suspiciousBody = 0;
     data->lastWasCR = 0;
 
+    /* Get my IP address */
+    me = smfi_getsymval(ctx, "{if_addr}");
+    if (me && *me && MyIPAddress && !strcmp(me, MyIPAddress)) {
+	data->myip = MyIPAddress;
+    } else if (me && *me && strcmp(me, "127.0.0.1")) {
+	data->myip = strdup_with_log(me);
+    } else {
+	/* Sigh... use our computed address */
+	data->myip = MyIPAddress;
+    }
+
+    /* Get my port */
+    if (!data->daemon_port) {
+	me = smfi_getsymval(ctx, "{daemon_port}");
+	if (me && *me) {
+	    sscanf(me, "%u", &(data->daemon_port));
+	}
+    }
+
+
     if (doRelayCheck) {
 	char buf2[SMALLBUF];
 	int n = MXRelayOK(MultiplexorSocketName, buf2, data->hostip,
-			  data->hostname);
+			  data->hostname, data->hostport, data->myip, data->daemon_port);
 	if (n == MD_REJECT) {
 	    /* Can't call smfi_setreply from connect callback */
 	    /* set_dsn(ctx, buf2, 5); */
@@ -727,7 +771,7 @@ helo(SMFICTX *ctx, char *helohost)
     if (doHeloCheck) {
 	char buf2[SMALLBUF];
 	int n = MXHeloOK(MultiplexorSocketName, buf2, data->hostip,
-			 data->hostname, data->heloArg);
+			 data->hostname, data->heloArg, data->hostport, data->myip, data->daemon_port);
 	if (n == MD_REJECT) {
 	    set_dsn(ctx, buf2, 5);
 	    cleanup(ctx);
@@ -780,7 +824,6 @@ envfrom(SMFICTX *ctx, char **from)
     char buffer[SMALLBUF];
     char buf2[SMALLBUF];
     char *queueid;
-    char *me;
     char **macro;
     dynamic_buffer dbuf;
 
@@ -907,6 +950,18 @@ envfrom(SMFICTX *ctx, char **from)
 	macro++;
     }
 
+    /* Fake client_port: We don't get the macro, but we have the connection
+       info cached in our private data area. */
+    dbuf_putc(&dbuf, '=');
+    append_percent_encoded(&dbuf, "client_port");
+    dbuf_putc(&dbuf, ' ');
+    {
+	char portstring[32];
+	snprintf(portstring, sizeof(portstring), "%u", data->hostport);
+	append_percent_encoded(&dbuf, portstring);
+    }
+    dbuf_putc(&dbuf, '\n');
+
     /* Write any additional macros requested by user */
     for (i=0; i<NumAdditionalMacros; i++) {
 	append_macro_value(&dbuf, ctx, AdditionalMacros[i]);
@@ -950,19 +1005,6 @@ envfrom(SMFICTX *ctx, char **from)
     dbuf_free(&dbuf);
     data->cmdFD = put_fd(data->cmdFD);
 
-
-    /* Get my IP address */
-    me = smfi_getsymval(ctx, "{if_addr}");
-    if (me && *me && MyIPAddress && !strcmp(me, MyIPAddress)) {
-	data->myip = MyIPAddress;
-    } else if (me && *me && strcmp(me, "127.0.0.1")) {
-	data->myip = strdup_with_log(me);
-    } else {
-	/* Sigh... use our computed address */
-	data->myip = MyIPAddress;
-    }
-
-
     if (doSenderCheck) {
 	int n = MXSenderOK(MultiplexorSocketName, buf2,
 			   (char const **) from, data->hostip, data->hostname,
@@ -1057,6 +1099,11 @@ static sfsistat mf_negotiate(SMFICTX *ctx,
 			     unsigned long *pf2,
 			     unsigned long *pf3)
 {
+    dynamic_buffer dbuf;
+    char **macroname;
+    int done_one = 0;
+    int i;
+
     *pf0 = f0;
     *pf1 = 0;
     *pf2 = 0;
@@ -1070,6 +1117,50 @@ static sfsistat mf_negotiate(SMFICTX *ctx,
     /* Don't want leading spaces */
     *pf1 &= (~SMFIP_HDR_LEADSPC);
 
+    /* Send along the list of macros we want */
+    dbuf_init(&dbuf);
+
+    for (macroname = StandardSendmailMacros; *macroname; macroname++) {
+	if (done_one) {
+	    dbuf_putc(&dbuf, ' ');
+	}
+	if (strlen(*macroname) > 1) {
+	    dbuf_putc(&dbuf, '{');
+	}
+	dbuf_puts(&dbuf, *macroname);
+	if (strlen(*macroname) > 1) {
+	    dbuf_putc(&dbuf, '}');
+	}
+	done_one = 1;
+    }
+
+    for (i=0, macroname=AdditionalMacros; i<NumAdditionalMacros; i++, macroname++) {
+	if (done_one) {
+	    dbuf_putc(&dbuf, ' ');
+	}
+	if (strlen(*macroname) > 1) {
+	    dbuf_putc(&dbuf, '{');
+	}
+	dbuf_puts(&dbuf, *macroname);
+	if (strlen(*macroname) > 1) {
+	    dbuf_putc(&dbuf, '}');
+	}
+	done_one = 1;
+    }
+
+    i = 1;
+    if (smfi_setsymlist(ctx, SMFIM_CONNECT, DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_HELO,    DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_ENVFROM, DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_ENVRCPT, DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_DATA,    DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_EOM,     DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+    if (smfi_setsymlist(ctx, SMFIM_EOH,     DBUF_VAL(&dbuf)) != MI_SUCCESS) i = 0;
+
+    dbuf_free(&dbuf);
+    if (!i) {
+	syslog(LOG_INFO, "smfi_setsymlist() failed");
+    }
     return SMFIS_CONTINUE;
 
 }
@@ -1614,7 +1705,7 @@ eom(SMFICTX *ctx)
     data->lastWasCR = 0;
 
     /* Run the filter */
-    if (MXScanDir(MultiplexorSocketName, data->dir) < 0) {
+    if (MXScanDir(MultiplexorSocketName, data->qid, data->dir) < 0) {
 	data->filterFailed = 1;
 	cleanup(ctx);
 	DEBUG_EXIT("eom", "SMFIS_TEMPFAIL");
@@ -1887,7 +1978,7 @@ eom(SMFICTX *ctx)
 		MD_SMFI_TRY(smfi_chgfrom, (ctx, rptr+1, NULL));
 	    }
 #else
-	    syslog(LOG_WARNING, "change_sender called, but this version of libmilter does not support CHGFROM");
+	    syslog(LOG_WARNING, "%s: change_sender called, but this version of libmilter does not support CHGFROM", data->qid);
 #endif
 	    break;
 	case 'S':
@@ -2162,12 +2253,14 @@ usage(void)
     fprintf(stderr, "  -H                -- Do HELO checks before processing any messages\n");
     fprintf(stderr, "  -c                -- Strip bare <CR> characters from message body\n");
     fprintf(stderr, "  -R num            -- Reserve num slaves for connections from localhost\n");
+    fprintf(stderr, "  -G                -- Make sockets group-writable and files group-readable\n");
     fprintf(stderr, "  -N                -- Do not pass rejected recipients to milter (Sendmail\n");
     fprintf(stderr, "                       8.14.0 and newer only.)\n");
 
     exit(EXIT_FAILURE);
 }
 
+#define REPORT_FAILURE(msg, len) do { if (kidpipe[1] >= 0) { write(kidpipe[1], "E" msg, len+1); } else { fprintf(stderr, "%s\n", msg); } } while(0)
 /**********************************************************************
 * %FUNCTION: main
 * %ARGUMENTS:
@@ -2190,6 +2283,11 @@ main(int argc, char **argv)
     int nodaemon = 0;
     char buf[SMALLBUF];
     int got_p_option = 0;
+    int kidpipe[2];
+    char kidmsg[256];
+
+    mode_t socket_umask = 077;
+    mode_t file_umask   = 077;
 
 #ifdef ENABLE_DEBUGGING
     /* Keep debugging malloc macros happy... */
@@ -2203,7 +2301,7 @@ main(int argc, char **argv)
     }
 
     /* Paranoia time */
-    umask(027);
+    umask(077);
 
     /* Paranoia time II */
     if (getuid() != geteuid()) {
@@ -2244,8 +2342,13 @@ main(int argc, char **argv)
     }
 
     /* Process command line options */
-    while ((c = getopt(argc, argv, "NCDHL:MP:R:S:TU:Xa:b:cdhkm:p:qrstvx:z:")) != -1) {
+    while ((c = getopt(argc, argv, "GNCDHL:MP:R:S:TU:Xa:b:cdhkm:p:qrstvx:z:")) != -1) {
 	switch (c) {
+	case 'G':
+	    socket_umask = 007;
+	    file_umask   = 027;
+	    break;
+
 	case 'N':
 #ifdef MILTER_BUILDLIB_HAS_NEGOTIATE
 	    SeeRejectedRecipients = 0;
@@ -2504,27 +2607,74 @@ main(int argc, char **argv)
 
     /* Daemonize */
     if (!nodaemon) {
+	/* Set up a pipe so child can report back when it's happy */
+	if (pipe(kidpipe) < 0) {
+	    perror("pipe");
+	    exit(EXIT_FAILURE);
+	}
+
 	i = fork();
 	if (i < 0) {
 	    fprintf(stderr, "%s: fork() failed\n", argv[0]);
 	    exit(EXIT_FAILURE);
 	} else if (i != 0) {
 	    /* parent */
-	    exit(EXIT_SUCCESS);
+	    close(kidpipe[1]);
+
+	    /* Wait for a message from kid */
+	    i = read(kidpipe[0], kidmsg, sizeof(kidmsg) - 1);
+	    if (i < 0) {
+		fprintf(stderr, "Error reading message from child: %s\n",
+			strerror(errno));
+		exit(1);
+	    }
+	    /* Zero-terminate the string */
+	    kidmsg[i] = 0;
+	    if (i == 1 && kidmsg[0] == 'X') {
+		/* Child indicated successful startup */
+		exit(EXIT_SUCCESS);
+	    }
+	    if (i > 1 && kidmsg[0] == 'E') {
+		/* Child indicated error */
+		fprintf(stderr, "Error from child: %s\n", kidmsg+1);
+		exit(EXIT_FAILURE);
+	    }
+	    /* Unknown status from child */
+	    fprintf(stderr, "Unknown reply from child: %s\n", kidmsg);
+	    exit(EXIT_FAILURE);
 	}
+
+	/* In the child */
+	close(kidpipe[0]);
 	setsid();
 	signal(SIGHUP, SIG_IGN);
 	i = fork();
 	if (i < 0) {
 	    fprintf(stderr, "%s: fork() failed\n", argv[0]);
+	    /* Signal the waiting parent */
+	    write(kidpipe[1], "Efork() failed", 14);
 	    exit(EXIT_FAILURE);
 	} else if (i != 0) {
 	    exit(EXIT_SUCCESS);
 	}
+    } else {
+	/* nodaemon */
+	kidpipe[0] = -1;
+	kidpipe[1] = -1;
+    }
+
+    /* Write pid */
+    if (pidfile) {
+	if (write_and_lock_pidfile(pidfile) < 0) {
+	    /* Signal the waiting parent */
+	    REPORT_FAILURE("Cannot lock pidfile: Is another copy running?", 45);
+	    exit(1);
+	}
+	free(pidfile);
     }
 
     (void) closelog();
-    closefiles();
+    closefiles(kidpipe[1]);
 
     /* Direct stdin/stdout/stderr to /dev/null */
     open("/dev/null", O_RDWR);
@@ -2532,13 +2682,16 @@ main(int argc, char **argv)
     open("/dev/null", O_RDWR);
 
     openlog("mimedefang", LOG_PID, facility);
-    /* Write pid */
-    if (pidfile) {
-	if (write_and_lock_pidfile(pidfile) < 0) {
-	    exit(1);
-	}
-	free(pidfile);
-    }
+
+    /* Open the milter socket if library has smfi_opensocket */
+#ifdef MILTER_BUILDLIB_HAS_OPENSOCKET
+    umask(socket_umask);
+    (void) smfi_opensocket(1);
+    umask(file_umask);
+#else
+    /* Gah, we can't create the socket, so use socket_umask throughout */
+    umask(socket_umask);
+#endif
 
     if (ValidateHeader[0]) {
 	syslog(LOG_DEBUG, "IP validation header is %s", ValidateHeader);
@@ -2550,19 +2703,31 @@ main(int argc, char **argv)
     signal(SIGBUS, handle_sig);
 #endif
 
-    /* Wait for the multiplexor to come alive */
+    /* Wait up to 20 seconds for the multiplexor to come alive */
     mx_alive = 0;
-    for (c=0; c<50; c++) {
+    for (c=0; c<100; c++) {
+	struct timeval sleeptime;
+	sleeptime.tv_sec = 0;
+	sleeptime.tv_usec = 200000;
 	if (MXCheckFreeSlaves(MultiplexorSocketName) >= 0) {
 	    mx_alive = 1;
 	    break;
 	}
-	sleep(3);
+	/* Sleep for 200ms */
+	select(0, NULL, NULL, NULL, &sleeptime);
     }
     if (mx_alive) {
 	syslog(LOG_INFO, "Multiplexor alive - entering main loop");
     } else {
-	syslog(LOG_WARNING, "Multiplexor unresponsive - entering main loop anyway");
+	/* Signal the waiting parent */
+	REPORT_FAILURE("Multiplexor socket did not appear.  Exiting.", 44);
+	exit(EXIT_FAILURE);
+    }
+
+    /* Tell the waiting parent that everything is A-OK */
+    if (kidpipe[1] >= 0) {
+	write(kidpipe[1], "X", 1);
+	close(kidpipe[1]);
     }
     return smfi_main();
 }
