@@ -26,12 +26,15 @@ extern int percent_encode(char const *in, char *out, int outlen);
 extern void percent_decode(char *buf);
 
 extern int MXCheckFreeSlaves(char const *sockname);
-extern int MXScanDir(char const *sockname, char const *dir);
+extern int MXScanDir(char const *sockname, char const *qid, char const *dir);
 extern int MXCommand(char const *sockname, char const *cmd, char *buf, int len);
 extern int MXRelayOK(char const *sockname, char *msg,
-		     char const *ip, char const *name);
+		     char const *ip, char const *name, unsigned int port,
+		     char const *myip, unsigned int daemon_port);
 extern int MXHeloOK(char const *sockname, char *msg,
-		    char const *helo, char const *ip, char const *name);
+		    char const *ip, char const *name, char const *helo,
+		    unsigned int port, char const *myip, unsigned int daemon_port);
+
 extern int MXSenderOK(char const *sockname, char *msg,
 		      char const **sender_argv, char const *ip, char const *name,
 		      char const *helo, char const *dir, char const *qid);
