@@ -12,9 +12,6 @@
 *
 ***********************************************************************/
 
-static char const RCSID[] =
-"$Id$";
-
 #include "config.h"
 #include "mimedefang.h"
 #include <sys/stat.h>
@@ -65,6 +62,7 @@ rm_r(void *ctx, char const *dir)
     struct dirent *entry;
     struct dirent *result = NULL;
     int n;
+    int retcode = 0;
 
     if (lstat(dir, &sbuf) < 0) {
 	syslog(LOG_WARNING, "lstat(%s) failed: %m", dir);
@@ -95,19 +93,21 @@ rm_r(void *ctx, char const *dir)
 
     d = opendir(dir);
     if (!d) {
+	int errno_save = errno;
 	syslog(LOG_WARNING, "opendir(%s) failed: %m", dir);
 	free(entry);
+	errno = errno_save;
 	return -1;
     }
 
     for (;;) {
 	n = readdir_r(d, entry, &result);
-	if (n < 0) {
-	    int errno_save = errno;
+	if (n != 0) {
+	    errno = n;
 	    syslog(LOG_WARNING, "readdir_r failed: %m");
 	    closedir(d);
 	    free(entry);
-	    errno = errno_save;
+	    errno = n;
 	    return -1;
 	}
 	if (!result) break;
@@ -117,11 +117,7 @@ rm_r(void *ctx, char const *dir)
 	}
 	snprintf(buf, sizeof(buf), "%s/%s", dir, entry->d_name);
 	if (rm_r(ctx, buf) < 0) {
-	    int errno_save = errno;
-	    closedir(d);
-	    free(entry);
-	    errno = errno_save;
-	    return -1;
+	    retcode = -1;
 	}
     }
     free(entry);
@@ -130,7 +126,7 @@ rm_r(void *ctx, char const *dir)
 	syslog(LOG_WARNING, "rmdir(%s) failed: %m", dir);
 	return -1;
     }
-    return 0;
+    return retcode;
 }
 
 #else /* EXEC_RM_FOR_CLEANUP */
@@ -145,7 +141,7 @@ rm_r(void *ctx, char const *dir)
 *  exec'ing "rm".
 ***********************************************************************/
 int
-rm_r(char const *dir)
+rm_r(void *ctx, char const *dir)
 {
     pid_t pid;
 
