@@ -12,9 +12,6 @@
 *
 ***********************************************************************/
 
-static char const RCSID[] =
-"$Id$";
-
 #define _BSD_SOURCE 1
 
 #ifdef __linux__
@@ -298,6 +295,15 @@ MXCommand(char const *sockname,
 	return MD_TEMPFAIL;
     }
     buf[nread] = 0;
+    /* If we read a full buffer, read to EOF to maintain synchronizaion */
+    if (nread == len-1) {
+	char slop[SMALLBUF];
+	if (readn(fd, slop, SMALLBUF) > 0) {
+	    syslog(LOG_WARNING, "MXCommand: Overlong reply from multiplexor was truncated!");
+	    /* Read all the way to EOF */
+	    while (readn(fd, slop, SMALLBUF) > 0);
+	}
+    }
     close(fd);
     return 0;
 }
@@ -328,6 +334,7 @@ MXCheckFreeSlaves(char const *sockname)
 * %FUNCTION: MXScanDir
 * %ARGUMENTS:
 *  sockname -- MX socket name
+*  qid -- Sendmail queue ID
 *  dir -- directory to scan
 * %RETURNS:
 *  0 if scanning succeeded; -1 if there was an error.
@@ -336,13 +343,21 @@ MXCheckFreeSlaves(char const *sockname)
 ***********************************************************************/
 int
 MXScanDir(char const *sockname,
+	  char const *qid,
 	  char const *dir)
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
     int len;
 
-    snprintf(cmd, SMALLBUF, "scan %s\n", dir);
+    if (!qid || !*qid) {
+	qid = "NOQUEUE";
+    }
+
+    if (percent_encode_command(1, cmd, sizeof(cmd), "scan", qid, dir, NULL) < 0) {
+	return MD_TEMPFAIL;
+    }
+
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return MD_TEMPFAIL;
 
     if (!strcmp(ans, "ok\n")) return 0;
@@ -550,6 +565,9 @@ munch_mx_return(char *ans, char *msg)
 *  msg -- buffer for holding error message, at least SMALLBUF chars
 *  ip -- relay IP address
 *  name -- relay name
+*  port -- client port
+*  myip -- My IP address, if known.
+*  daemon_port -- Listening port
 * %RETURNS:
 *  1 if it's OK to accept connections from this host; 0 if not, -1 if error.
 *  If connection is rejected, error message *may* be set.
@@ -558,11 +576,20 @@ int
 MXRelayOK(char const *sockname,
 	  char *msg,
 	  char const *ip,
-	  char const *name)
+	  char const *name,
+	  unsigned int port,
+	  char const *myip,
+	  unsigned int daemon_port)
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
 
+    char port_string[65];
+    char daemon_port_string[65];
+
+    snprintf(port_string, sizeof(port_string), "%u", port);
+    snprintf(daemon_port_string, sizeof(daemon_port_string), "%u", daemon_port);
+
     *msg = 0;
 
     if (!ip || !*ip) {
@@ -571,7 +598,11 @@ MXRelayOK(char const *sockname,
     if (!name || !*name) {
 	name = ip;
     }
-    if (percent_encode_command(1, cmd, sizeof(cmd), "relayok", ip, name, NULL) < 0) {
+    if (!myip || !*myip) {
+	myip = "UNKNOWN";
+    }
+
+    if (percent_encode_command(1, cmd, sizeof(cmd), "relayok", ip, name, port_string, myip, daemon_port_string, NULL) < 0) {
 	return MD_TEMPFAIL;
     }
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return MD_TEMPFAIL;
@@ -583,7 +614,12 @@ MXRelayOK(char const *sockname,
 * %ARGUMENTS:
 *  sockname -- multiplexor socket name
 *  msg -- buffer for holding error message, at least SMALLBUF chars
+*  ip -- IP address of client
+*  name -- resolved name of client
 *  helo -- the helo string
+*  port -- client port
+*  myip -- My IP address, if known.
+*  daemon_port -- Listening port
 * %RETURNS:
 *  1 if it's OK to accept messages from this sender; 0 if not, -1 if error or
 *  we should tempfail.
@@ -593,11 +629,20 @@ MXHeloOK(char const *sockname,
 	 char *msg,
 	 char const *ip,
 	 char const *name,
-	 char const *helo)
+	 char const *helo,
+	 unsigned int port,
+	 char const *myip,
+	 unsigned int daemon_port)
 {
     char cmd[SMALLBUF];
     char ans[SMALLBUF];
 
+    char port_string[65];
+    char daemon_port_string[65];
+
+    snprintf(port_string, sizeof(port_string), "%u", port);
+    snprintf(daemon_port_string, sizeof(daemon_port_string), "%u", daemon_port);
+
     *msg = 0;
 
     if (!ip || !*ip) {
@@ -609,7 +654,12 @@ MXHeloOK(char const *sockname,
     if (!helo) {
 	helo = "UNKNOWN";
     }
-    if (percent_encode_command(1, cmd, sizeof(cmd), "helook", ip, name, helo, NULL) < 0) {
+
+    if (!myip || !*myip) {
+	myip = "UNKNOWN";
+    }
+
+    if (percent_encode_command(1, cmd, sizeof(cmd), "helook", ip, name, helo, port_string, myip, daemon_port_string, NULL) < 0) {
 	return MD_TEMPFAIL;
     }
     if (MXCommand(sockname, cmd, ans, SMALLBUF-1) < 0) return MD_TEMPFAIL;
@@ -665,6 +715,7 @@ MXSenderOK(char const *sockname,
     if (!helo) {
 	helo = "UNKNOWN";
     }
+
     if (percent_encode_command(0, cmd, sizeof(cmd)-1, "senderok", sender, ip,
 			       name,
 			       helo, dir, qid, NULL) < 0) {
